"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.parser.LipidBaseParserEventHandler import LipidBaseParserEventHandler
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidClass import *
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.Adduct import Adduct

from pygoslin.domain.LipidCompleteStructure import LipidCompleteStructure
from pygoslin.domain.LipidFullStructure import LipidFullStructure
from pygoslin.domain.LipidStructureDefined import LipidStructureDefined
from pygoslin.domain.LipidSnPosition import LipidSnPosition
from pygoslin.domain.LipidMolecularSpecies import LipidMolecularSpecies
from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.Cycle import Cycle
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo

from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.HeadGroup import HeadGroup
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidClass import *


class LipidMapsParserEventHandler(LipidBaseParserEventHandler):
    def __init__(self):
        super().__init__()
        self.reset_lipid(None)
        
        self.registered_events["lipid_pre_event"] = self.reset_lipid
        self.registered_events["lipid_post_event"] = self.build_lipid
        
        ## set adduct events
        self.registered_events["adduct_info_pre_event"] = self.new_adduct
        self.registered_events["adduct_pre_event"] = self.add_adduct
        self.registered_events["charge_pre_event"] = self.add_charge
        self.registered_events["charge_sign_pre_event"] = self.add_charge_sign
        
        self.registered_events["mediator_pre_event"] = self.mediator_event
        self.registered_events["fa_no_hg_pre_event"] = self.pure_fa
        
        self.registered_events["sgl_species_pre_event"] = self.set_species_level
        self.registered_events["tgl_species_pre_event"] = self.set_species_level
        self.registered_events["dpl_species_pre_event"] = self.set_species_level
        self.registered_events["cl_species_pre_event"] = self.set_species_level
        self.registered_events["dsl_species_pre_event"] = self.set_species_level
        self.registered_events["species_fa_pre_event"] = self.set_species_level
        self.registered_events["fa2_unsorted_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["fa3_unsorted_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["fa4_unsorted_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["hg_dg_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["fa_lpl_molecular_pre_event"] = self.set_molecular_subspecies_level
        self.registered_events["hg_lbpa_pre_event"] = self.set_molecular_subspecies_level
        
        self.registered_events["hg_sgl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_gl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_cl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_dpl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_lpl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_fourpl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_threepl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_dsl_pre_event"] = self.set_head_group_name
        self.registered_events["hg_lsl_pre_event"] = self.set_head_group_name
        self.registered_events["ch_pre_event"] = self.set_head_group_name
        self.registered_events["pk_hg_pre_event"] = self.set_head_group_name
        self.registered_events["hg_fa_pre_event"] = self.set_head_group_name
        self.registered_events["hg_cpa_pre_event"] = self.set_head_group_name
        self.registered_events["hg_che_pre_event"] = self.set_head_group_name
        self.registered_events["special_cer_pre_event"] = self.set_head_group_name
        self.registered_events["special_cer_hg_pre_event"] = self.set_head_group_name
        self.registered_events["omega_linoleoyloxy_Cer_pre_event"] = self.set_omega_head_group_name
        
        self.registered_events["glyco_struct_pre_event"] = self.add_glyco
        
        self.registered_events["lcb_pre_event"] = self.new_lcb
        self.registered_events["lcb_post_event"] = self.clean_lcb
        self.registered_events["fa_pre_event"] = self.new_fa
        self.registered_events["fa_post_event"] = self.append_fa
        self.registered_events["c_type_pre_event"] = self.c_type
        self.registered_events["c_type_post_event"] = self.append_fa
        
        self.registered_events["db_single_position_pre_event"] = self.set_isomeric_level
        self.registered_events["db_single_position_post_event"] = self.add_db_position
        self.registered_events["db_position_number_pre_event"] = self.add_db_position_number
        self.registered_events["cistrans_pre_event"] = self.add_cistrans
        
        self.registered_events["ether_prefix_pre_event"] = self.add_ether
        self.registered_events["ether_suffix_pre_event"] = self.add_ether
        self.registered_events["lcb_pure_fa_pre_event"] = self.add_dihydroxyl
        self.registered_events["hydroxyl_pre_event"] = self.add_hydroxyl
        self.registered_events["hydroxyl_lcb_pre_event"] = self.add_hydroxyl_lcb
        self.registered_events["db_count_pre_event"] = self.add_double_bonds
        self.registered_events["carbon_pre_event"] = self.add_carbon
    
        self.registered_events["structural_mod_pre_event"] = self.set_structural_subspecies_level
        self.registered_events["single_mod_pre_event"] = self.set_mod
        self.registered_events["mod_text_pre_event"] = self.set_mod_text
        self.registered_events["mod_pos_pre_event"] = self.set_mod_pos
        self.registered_events["mod_num_pre_event"] = self.set_mod_num
        self.registered_events["single_mod_post_event"] = self.add_functional_group
        self.registered_events["additional_modifier_pre_event"] = self.add_additional_modifier
        
        self.registered_events["special_cer_prefix_pre_event"] = self.add_ACer
        
        self.registered_events["isotope_pair_pre_event"] = self.new_adduct
        self.registered_events["isotope_element_pre_event"] = self.set_heavy_element
        self.registered_events["isotope_number_pre_event"] = self.set_heavy_number
        
        self.registered_events["sphinga_pre_event"] = self.new_sphinga
        self.registered_events["hg_lsl_pure_pre_event"] = self.new_sph
        self.registered_events["sphinga_phospho_pre_event"] = self.add_phospho
        self.registered_events["sphinga_suffix_pre_event"] = self.sphinga_db_set
        self.registered_events["sphinga_lcb_len_pre_event"] = self.add_carbon_pre_len
        self.registered_events["sphinga_prefix_pre_event"] = self.set_hydro_pre_num
        self.registered_events["sphinga_hg_pure_pre_event"] = self.new_sphinga_pure
        self.registered_events["sphinga_hg_pure_post_event"] = self.clean_lcb
        
        
        
        self.debug = ""
        
        
        
    def reset_lipid(self, node):
        self.level = LipidLevel.FULL_STRUCTURE
        self.adduct = None
        self.head_group = ""
        self.lcb = None
        self.fa_list = []
        self.current_fa = None
        self.use_head_group = False
        self.omit_fa = False
        self.db_position = 0
        self.db_cistrans = ""
        self.db_numbers = -1
        self.mod_text = ""
        self.mod_pos = -1
        self.mod_num = 1
        self.headgroup_decorators = []
        self.add_omega_linoleoyloxy_Cer = False
        self.heavy_number = 0
        self.heavy_element = None
        self.sphinga_pure = False
        self.lcb_carbon_pre_set = 18
        self.lcb_db_pre_set = 0
        self.lcb_hydro_pre_set = []
        self.sphinga_prefix = ""
        self.sphinga_suffix = ""
        
        
        
    def set_molecular_subspecies_level(self, node):
        self.set_lipid_level(LipidLevel.MOLECULAR_SPECIES)
        
        
        
    def add_carbon_pre_len(self, node):
        self.lcb_carbon_pre_set = int(node.get_text())
        
        
        
    def mediator_event(self, node):
        self.use_head_group = True
        self.head_group = node.get_text()
        
        
        
    def sphinga_db_set(self, node):
        self.sphinga_suffix = node.get_text()
        if self.sphinga_suffix == 'anine': self.lcb_db_pre_set = 0
        elif self.sphinga_suffix == 'osine': self.lcb_db_pre_set = 1
        elif self.sphinga_suffix == 'adienine': self.lcb_db_pre_set = 2
        
        
    def c_type(self, node):
        if self.head_group == "Cer":
            self.new_sphinga_pure(node)
        else:
            self.sphinga_pure = True
            self.lcb_hydro_pre_set = [get_functional_group("OH")]
            self.lcb_hydro_pre_set[0].position = 3
            self.new_lcb(node)
        self.clean_lcb(node)
        self.new_fa(node)
        
        
    def new_sphinga(self, node):
        self.head_group = "SPB"
        
        
    def new_sph(self, node):
        if self.head_group.lower() == "sph":
            self.new_sphinga_pure(node)
        else:
            self.sphinga_pure = True
            self.lcb_hydro_pre_set = [get_functional_group("OH")]
            self.lcb_hydro_pre_set[0].position = 3
            self.new_lcb(node)
        self.clean_lcb(node)
        
        
    def new_sphinga_pure(self, node):
        self.sphinga_pure = True
        self.lcb_hydro_pre_set = [get_functional_group("OH"), get_functional_group("OH")]
        self.lcb_hydro_pre_set[0].position = 1
        self.lcb_hydro_pre_set[1].position = 3
        self.new_lcb(node)
        
        
        
    def set_hydro_pre_num(self, node):
        self.lcb_hydro_pre_set.append(get_functional_group("OH"))
        self.lcb_hydro_pre_set[-1].position = 4
        self.sphinga_prefix = node.get_text()
        
        
        
    def add_phospho(self, node):
        phospho_suffix = node.get_text()
        if phospho_suffix == "1-phosphate":
            self.head_group += "P"
        elif phospho_suffix == "1-phosphocholine":
            self.head_group = "LSM"
        self.lcb_hydro_pre_set = self.lcb_hydro_pre_set[1:]

        
        
    def set_isomeric_level(self, node):
        self.db_position = 0
        self.db_cistrans = ""
        
        
        
    def set_heavy_element(self, node):
        self.adduct.heavy_elements[Element.H2] = 0
        
        
        
    def set_heavy_number(self, node):
        self.adduct.heavy_elements[Element.H2] = int(node.get_text())
        
    
    
    def add_db_position(self, node):
        if self.current_fa != None:
            if type(self.current_fa.double_bonds) == int:
                self.db_numbers = self.current_fa.double_bonds
                self.current_fa.double_bonds = {self.db_position: self.db_cistrans}
            else:
                self.current_fa.double_bonds[self.db_position] = self.db_cistrans
            if self.db_cistrans not in {"E", "Z"}: self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
    
    acer_heads = {"1-O-myristoyl": FattyAcid("FA", 14),
                  "1-O-palmitoyl": FattyAcid("FA", 16),
                  "1-O-stearoyl": FattyAcid("FA", 18),
                  "1-O-eicosanoyl": FattyAcid("FA", 20),
                  "1-O-behenoyl": FattyAcid("FA", 22),
                  "1-O-lignoceroyl": FattyAcid("FA", 24),
                  "1-O-cerotoyl": FattyAcid("FA", 26),
                  "1-O-pentacosanoyl": FattyAcid("FA", 25),
                  "1-O-carboceroyl": FattyAcid("FA", 28),
                  "1-O-tricosanoyl": FattyAcid("FA", 30),
                  "1-O-lignoceroyl-omega-linoleoyloxy": FattyAcid("FA", 24),
                  "1-O-stearoyl-omega-linoleoyloxy": FattyAcid("FA", 18)
                  }
                  
        
        
    def add_ACer(self, node):
        head = node.get_text()
        self.head_group = "ACer"
        
        if head not in LipidMapsParserEventHandler.acer_heads:
            raise LipidException("ACer head group '%s' unknown" % head)
        
        hgd = HeadgroupDecorator("decorator_acyl", suffix = True)
        hgd.functional_groups["decorator_acyl"] = [LipidMapsParserEventHandler.acer_heads[head]]
        self.headgroup_decorators.append(hgd)
        
        if head in {"1-O-lignoceroyl-omega-linoleoyloxy", "1-O-stearoyl-omega-linoleoyloxy"}:
            self.add_omega_linoleoyloxy_Cer = True
        
        
        
    def add_db_position_number(self, node):
        self.db_position = int(node.get_text())
        
        

    def add_cistrans(self, node):
        self.db_cistrans = node.get_text()
        
        
        
    def pure_fa(self, node):
        self.head_group = "FA"
        
        
        
    def set_head_group_name(self, node):
        if len(self.head_group) == 0:
            self.head_group = node.get_text()
        
        
        
    def set_omega_head_group_name(self, node):
        self.set_head_group_name(node)
        self.add_omega_linoleoyloxy_Cer = True
        
        
        
    def add_additional_modifier(self, node):
        modifier = node.get_text()
        if modifier == "h":
            functional_group = get_functional_group("OH")
            self.current_fa.add_functional_group(functional_group)
            self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)
        
        
        
    def set_species_level(self, node):
        self.set_lipid_level(LipidLevel.SPECIES)
        
        
        
    def set_structural_subspecies_level(self, node):
        self.set_lipid_level(LipidLevel.STRUCTURE_DEFINED)



    def set_mod(self, node):
        self.mod_text = ""
        self.mod_pos = -1
        self.mod_num = 1



    def set_mod_text(self, node):
        self.mod_text = node.get_text()
        
        
        
    def add_glyco(self, node):
        glyco = node.get_text()
        try:
            functional_group = get_functional_group(glyco)
        except Exception:
            raise LipidParsingException("Carbohydrate '%s' unknown" % glyco)
        
        functional_group.elements[Element.O] -= 1
        self.headgroup_decorators.append(functional_group)
        


    def set_mod_pos(self, node):
        self.mod_pos = int(node.get_text())


    def set_mod_num(self, node):
        self.mod_num = int(node.get_text())
        
        
        
    def add_functional_group(self, node):
        if self.mod_text != "Cp":
            if self.current_fa.lipid_FA_bond_type in FattyAcid.LCB_STATES and self.mod_text == "OH" and "OH" in self.current_fa.functional_groups and len(self.current_fa.functional_groups["OH"]) > 0:
                    self.current_fa.functional_groups["OH"][-1].position = self.mod_pos
            else:
                functional_group = get_functional_group(self.mod_text)
                functional_group.position = self.mod_pos
                functional_group.count = self.mod_num
                self.current_fa.add_functional_group(functional_group)
        else:
            self.current_fa.num_carbon += 1
            self.current_fa.add_functional_group(Cycle(3, start = self.mod_pos, end = self.mod_pos + 2))
          
          
          
    def new_fa(self, node):
        self.db_numbers = -1
        self.current_fa = FattyAcid("FA")

        

    def append_fa(self, node):
        if type(self.current_fa.double_bonds) != int:
            if self.db_numbers > -1 and self.db_numbers != len(self.current_fa.double_bonds):
                raise LipidException("Double bond count does not match with number of double bond positions")
        elif self.current_fa.double_bonds > 0:
                self.set_lipid_level(LipidLevel.SN_POSITION)
            
        
        if self.level in {LipidLevel.SN_POSITION, LipidLevel.STRUCTURE_DEFINED, LipidLevel.FULL_STRUCTURE, LipidLevel.COMPLETE_STRUCTURE}:
            self.current_fa.position = len(self.fa_list) + 1
            
        if self.current_fa.num_carbon == 0: self.omit_fa = True
        self.fa_list.append(self.current_fa)
        
        self.current_fa = None
        
        
        
    def new_lcb(self, node):
        self.lcb = FattyAcid("LCB")
        self.current_fa = self.lcb
        self.lcb.set_type(LipidFaBondType.LCB_REGULAR)
        
            
            
    def clean_lcb(self, node):
        if self.sphinga_pure:
            self.lcb.num_carbon = self.lcb_carbon_pre_set
            self.lcb.double_bonds = self.lcb_db_pre_set
            self.lcb.functional_groups["OH"] = self.lcb_hydro_pre_set
        
        if self.sphinga_suffix != "":
            if (self.sphinga_suffix == "anine" and self.lcb.db_num() != 0) or (self.sphinga_suffix == "osine" and self.lcb.db_num() != 1) or (self.sphinga_suffix == "adienine" and self.lcb.db_num() != 2):
                raise LipidException("Double bond count does not match with head group description")
            
        if self.sphinga_prefix == "Phyto" and not self.sphinga_pure:
            pos_hydro = {fg.position for fg in self.lcb.functional_groups["OH"]}
            if len(self.lcb.functional_groups) == 0 or "OH" not in self.lcb.functional_groups or 4 not in pos_hydro:
                raise LipidException("hydroxyl count does not match with head group description")
            
        if type(self.current_fa.double_bonds) != int:
            if self.db_numbers > -1 and self.db_numbers != len(self.current_fa.double_bonds):
                raise LipidException("Double bond count does not match with number of double bond positions")
        elif self.current_fa.double_bonds > 0:
            self.set_lipid_level(LipidLevel.SN_POSITION)
        if "OH" in self.current_fa.functional_groups:
            for fg in self.current_fa.functional_groups["OH"]:
                if fg.position < 1:
                    self.set_structural_subspecies_level(node)
                    break
        self.current_fa = None
        
        
        
    def add_ether(self, node):
        ether = node.get_text()
        
        if ether in {"O-", "e"}: self.current_fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMANYL
        elif ether in {"P-", "p"}: self.current_fa.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMENYL
        
        
        
    def add_hydroxyl(self, node):
        num_h = int(node.get_text())
        if num_h <= 0: return
        functional_group = get_functional_group("OH")
        
        if self.sp_regular_lcb(): num_h -= 1
        
        functional_group.count = num_h
        if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
        self.current_fa.functional_groups["OH"].append(functional_group)
        
        
        
    def add_dihydroxyl(self, node):
        functional_group_p3 = get_functional_group("OH")
        functional_group_p3.position = 3
        
        if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
        self.current_fa.functional_groups["OH"].append(functional_group_p3)
        
        if not self.sp_regular_lcb():
            functional_group_p1 = get_functional_group("OH")
            functional_group_p1.position = 1
            self.current_fa.functional_groups["OH"].append(functional_group_p1)
        
        
        
    def add_hydroxyl_lcb(self, node):
        hydroxyl = node.get_text()
        if "OH" not in self.current_fa.functional_groups: self.current_fa.functional_groups["OH"] = []
        
        
        if hydroxyl == "m":
            functional_group_p3 = get_functional_group("OH")
            functional_group_p3.position = 3
            self.current_fa.functional_groups["OH"].append(functional_group_p3)
            
        elif hydroxyl == "d":
            if not self.sp_regular_lcb():
                functional_group_p1 = get_functional_group("OH")
                functional_group_p1.position = 1
                self.current_fa.functional_groups["OH"].append(functional_group_p1)
            
            functional_group_p3 = get_functional_group("OH")
            functional_group_p3.position = 3
            self.current_fa.functional_groups["OH"].append(functional_group_p3)
            
        elif hydroxyl == "t":
            if not self.sp_regular_lcb():
                functional_group_p1 = get_functional_group("OH")
                functional_group_p1.position = 1
                self.current_fa.functional_groups["OH"].append(functional_group_p1)
            
            functional_group_p3 = get_functional_group("OH")
            functional_group_p3.position = 3
            self.current_fa.functional_groups["OH"].append(functional_group_p3)
            
            functional_group_t = get_functional_group("OH")
            functional_group_t.position = 4
            self.current_fa.functional_groups["OH"].append(functional_group_t)
            
        
        
    def add_double_bonds(self, node):
        self.current_fa.double_bonds += int(node.get_text())
        
        
        
    def add_carbon(self, node):
        self.current_fa.num_carbon = int(node.get_text())
        
        

    def build_lipid(self, node):
        if self.lcb != None:
            self.fa_list = [self.lcb] + self.fa_list
            
        if self.add_omega_linoleoyloxy_Cer:
            if len(self.fa_list) != 2:
                raise LipidException("omega-linoleoyloxy-Cer with a different combination to one long chain base and one fatty acyl chain unknown")
            if "acyl" not in self.fa_list[-1].functional_groups: self.fa_list[-1].functional_groups["acyl"] = []
            self.fa_list[-1].functional_groups["acyl"].append(AcylAlkylGroup(FattyAcid("FA", 18, {9: "Z", 12: "Z"})))
            self.head_group = "Cer"
        
        headgroup = self.prepare_headgroup_and_checks()

        lipid = LipidAdduct()
        lipid.lipid = self.assemble_lipid(headgroup)
        lipid.adduct = self.adduct
        
        self.content = lipid
        
        
        
    def new_adduct(self, node):
        if self.adduct == None: self.adduct = Adduct("", "")
        
        
        
    def add_adduct(self, node):
        self.adduct.adduct_string = node.get_text()
        
        
        
    def add_charge(self, node):
        self.adduct.charge = int (node.get_text())
        
        
        
    def add_charge_sign(self, node):
        sign = node.get_text()
        if sign == "+": self.adduct.set_charge_sign(1)
        if sign == "-": self.adduct.set_charge_sign(-1)
        if self.adduct.charge == 0: self.adduct.charge = 1
        
        



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.FunctionalGroup import FunctionalGroup
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidClass import *
from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.LipidAdduct import LipidAdduct

class LipidMolecularSpecies(LipidSpecies):
    def __init__(self, head_group, fa = []):
        super().__init__(head_group, fa)
        self.fa = {}
        
        self.info.level = LipidLevel.MOLECULAR_SPECIES
        
        for fatty_acid in fa:
            if fatty_acid.name in self.fa:
                raise ConstraintViolationException("FA names must be unique! FA with name %s was already added!" % fatty_acid.name)
            
            self.fa[fatty_acid.name] = fatty_acid
            self.fa_list.append(fatty_acid)
                
                
        # add 0:0 dummys
        for i in range(len(fa), self.info.total_fa):
            fatty_acid = FattyAcid("FA%i" % (i + 1), position = -1)
            fatty_acid.unresolved_hidden_fa = 1 < len(fa) < self.info.poss_fa
            self.info.add(fatty_acid)
            self.fa[fatty_acid.name] = fatty_acid
            self.fa_list.append(fatty_acid)
    

    def get_extended_class(self):
        return super().get_extended_class()
    
    

    def build_lipid_subspecies_name(self, level):
        if level == None: level = LipidLevel.MOLECULAR_SPECIES
        fa_separator = "/" if level != LipidLevel.MOLECULAR_SPECIES or all_lipids[self.headgroup.lipid_class]["category"] == LipidCategory.SP else "_"

        lipid_name = [self.headgroup.get_lipid_string(level)]

        fa_headgroup_separator = " " if all_lipids[self.headgroup.lipid_class]["category"] != LipidCategory.ST else "/"
        
        if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE, LipidLevel.STRUCTURE_DEFINED, LipidLevel.SN_POSITION}:
            fa_string = fa_separator.join(fatty_acid.to_string(level) for fatty_acid in self.fa_list)
            if len(fa_string) > 0: lipid_name += [fa_headgroup_separator, fa_string]
        else:
            fa_string = fa_separator.join(fatty_acid.to_string(level) for fatty_acid in self.fa_list if fatty_acid.num_carbon > 0)
            if len(fa_string) > 0: lipid_name += [fa_headgroup_separator, fa_string]
            
        return "".join(lipid_name)
    
    
    
    def get_elements(self):
        dummy = FunctionalGroup("headgroup", elements = self.headgroup.get_elements())
        
        # add elements from all fatty acyl chains
        for fa in self.fa_list: dummy += fa
        
        return dummy.elements
    
    
    def get_lipid_string(self, level = None):
        if level == None or level == LipidLevel.MOLECULAR_SPECIES:
            return self.build_lipid_subspecies_name(LipidLevel.MOLECULAR_SPECIES)
        
        elif level in (LipidLevel.CATEGORY, LipidLevel.CLASS, LipidLevel.SPECIES):
            return super().get_lipid_string(level)
        else:
            raise Exception("LipidMolecularSpecies does not know how to create a lipid string for level %s" % level)
    
    
    def sort_fatty_acyl_chains(self):
        if self.info.level != LipidLevel.MOLECULAR_SPECIES and len(self.fa_list) < 2: return
        self.fa_list.sort(key = lambda fa: (fa.num_carbon != 0, fa.lipid_FA_bond_type.value, fa.num_carbon, fa.get_double_bonds(), LipidAdduct.compute_mass(fa.get_elements())))



from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.Element import Element
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.Cycle import Cycle
from pygoslin.domain.LipidAdduct import LipidAdduct
from pygoslin.domain.LipidSpecies import LipidSpecies
from pygoslin.domain.LipidLevel import LipidLevel
import os
from pygoslin.parser.Parser import *
import time
import csv

import unittest


class ShorthandTest(unittest.TestCase):
    
    def test_sphingolipids(self):
        parser = ShorthandParser()
        
        l = parser.parse("Cer 18:1(8Z);1OH,3OH/24:0")
        self.assertEqual(l.get_lipid_string(), "Cer 18:1(8Z);1OH,3OH/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "Cer 18:1(8);(OH)2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "Cer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C42H83NO3")
        
        l = parser.parse("Cer 18:1(8);(OH)2/24:0")
        self.assertEqual(l.get_lipid_string(), "Cer 18:1(8);(OH)2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "Cer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C42H83NO3")
        
        l = parser.parse("Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "Cer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C42H83NO3")
        
        l = parser.parse("Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "Cer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "Cer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C42H83NO3")
        
        l = parser.parse("Cer 42:1;O2")
        self.assertEqual(l.get_lipid_string(), "Cer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C42H83NO3")
        
        
        l = parser.parse("cLPA 18:0")
        self.assertEqual(l.get_lipid_string(), "CPA 18:0")
        self.assertEqual(l.get_sum_formula(), "C21H41O6P")
        
        
        l = parser.parse("Gal-Cer(1) 18:1(5Z);3OH/24:0")
        self.assertEqual(l.get_lipid_string(), "Gal-Cer(1) 18:1(5Z);3OH/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "Gal-Cer 18:1(5);OH/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "HexCer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C48H93NO8")
        
        l = parser.parse("Gal-Cer 18:1(5);OH/24:0")
        self.assertEqual(l.get_lipid_string(), "Gal-Cer 18:1(5);OH/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "HexCer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C48H93NO8")
        
        l = parser.parse("GalCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "HexCer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C48H93NO8")
        
        l = parser.parse("HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(), "HexCer 18:1;O2/24:0")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "HexCer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C48H93NO8")
        
        l = parser.parse("GalCer 42:1;O2")
        self.assertEqual(l.get_lipid_string(), "HexCer 42:1;O2")
        self.assertEqual(l.get_sum_formula(), "C48H93NO8")
        
        l = parser.parse("LSM 18:1;O2")
        self.assertEqual(l.get_lipid_string(), "LSM 18:1;O2")
        
        l = parser.parse("LSM(1) 18:1(5Z);3OH")
        self.assertEqual(l.get_lipid_string(), "LSM(1) 18:1(5Z);3OH")
        
        l = parser.parse("LHexCer 18:1;O2")
        self.assertEqual(l.get_lipid_string(), "LHexCer 18:1;O2")
        
        l = parser.parse("LHexCer 18:1;O2/0:0")
        self.assertEqual(l.get_lipid_string(), "LHexCer 18:1;O2")
        
        l = parser.parse("LHexCer(1) 18:1(5E);3OH/0:0")
        self.assertEqual(l.get_lipid_string(), "LHexCer(1) 18:1(5E);3OH")
        
        
        
        l = parser.parse("SPB 18:1(4Z);1OH,3OH")
        self.assertEqual(l.get_lipid_string(), "SPB 18:1(4Z);1OH,3OH")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "SPB 18:1(4);(OH)2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_sum_formula(), "C18H37NO2")
        
        l = parser.parse("SPB 18:1(4);(OH)2")
        self.assertEqual(l.get_lipid_string(), "SPB 18:1(4);(OH)2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_sum_formula(), "C18H37NO2")
        
        l = parser.parse("SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "SPB 18:1;O2")
        self.assertEqual(l.get_sum_formula(), "C18H37NO2")
        

        
        l = parser.parse("LSM(1) 17:1(4E);3OH")
        self.assertEqual(l.get_lipid_string(), "LSM(1) 17:1(4E);3OH")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "LSM 17:1(4);OH")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_sum_formula(), "C22H47N2O5P")
        
        l = parser.parse("LSM 17:1(4);OH")
        self.assertEqual(l.get_lipid_string(), "LSM 17:1(4);OH")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_sum_formula(), "C22H47N2O5P")
        
        l = parser.parse("LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "LSM 17:1;O2")
        self.assertEqual(l.get_sum_formula(), "C22H47N2O5P")
        

        
        l = parser.parse("EPC(1) 14:1(4E);3OH/20:1(11Z)")
        self.assertEqual(l.get_lipid_string(), "EPC(1) 14:1(4E);3OH/20:1(11Z)")
        self.assertEqual(l.get_lipid_string(LipidLevel.STRUCTURE_DEFINED), "EPC 14:1(4);OH/20:1(11)")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "EPC 34:2;O2")
        self.assertEqual(l.get_sum_formula(), "C36H71N2O6P")
        
        l = parser.parse("EPC 14:1(4);OH/20:1(11)")
        self.assertEqual(l.get_lipid_string(), "EPC 14:1(4);OH/20:1(11)")
        self.assertEqual(l.get_lipid_string(LipidLevel.SN_POSITION), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "EPC 34:2;O2")
        self.assertEqual(l.get_sum_formula(), "C36H71N2O6P")
        
        l = parser.parse("EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.MOLECULAR_SPECIES), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "EPC 34:2;O2")
        self.assertEqual(l.get_sum_formula(), "C36H71N2O6P")
        
        l = parser.parse("EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(), "EPC 14:1;O2/20:1")
        self.assertEqual(l.get_lipid_string(LipidLevel.SPECIES), "EPC 34:2;O2")
        self.assertEqual(l.get_sum_formula(), "C36H71N2O6P")
        
        l = parser.parse("EPC 34:2;O2")
        self.assertEqual(l.get_lipid_string(), "EPC 34:2;O2")
        self.assertEqual(l.get_sum_formula(), "C36H71N2O6P")

        try:
            l = parser.parse("SM 21:0;2O/21:1(3E)")
            self.assertTrue(False)
        except AssertionError:
            self.assertTrue(False, "SM 21:0;2O/21:1(3E) was excepted")
        except Exception:
            pass
        
        
    def test_nomenclature(self):
        data = []
        file_name = os.path.join("pygoslin", "data", "goslin", "testfiles", "shorthand-test.csv")
        with open(file_name, mode = "rt", encoding= "utf-8") as infile:
            lipidreader = csv.reader(infile, delimiter=',', quotechar='"')
            for row in lipidreader:
                row[-1] = row[-1].strip(" ")
                data.append(row)
                
        parser = ShorthandParser()
        
        
        levels = [LipidLevel.FULL_STRUCTURE, LipidLevel.STRUCTURE_DEFINED, LipidLevel.SN_POSITION, LipidLevel.MOLECULAR_SPECIES, LipidLevel.SPECIES]
        for row in data:
            
            lipid_name = row[0]
            lipid = parser.parse(lipid_name)
            formula = row[len(levels)] if len(row) > len(levels) else lipid.get_sum_formula()
            
            if len(row) > len(levels) and len(row[len(levels)]) > 0:
                formula = row[len(levels)]
                self.assertEqual(formula, lipid.get_sum_formula(), "test on lipid '%s'" % lipid_name)
            else:
                formula = lipid.get_sum_formula()
            
            for l, lipid_level in enumerate(levels):
                n = lipid.get_lipid_string(lipid_level)
                self.assertEqual(row[l], n, "first test on lipid '%s' and level '%s'" % (lipid_name, lipid_level))
                self.assertEqual(formula, lipid.get_sum_formula(), "first test on lipid '%s' and level '%s'" % (lipid_name, lipid_level))


                lipid2 = parser.parse(n)
                for ll in range(l, len(levels)):
                    self.assertEqual(row[ll], lipid2.get_lipid_string(levels[ll]), "second test on lipid '%s' and level '%s'" % (lipid_name, lipid_level))
                    self.assertEqual(formula, lipid2.get_sum_formula(), "second test on lipid '%s' and level '%s'" % (n, lipid_level))
        
        
    """
    def test_performance(self):
        cycles, parser = 50, ShorthandParser()
        length = 0
        
        start = time.time()
        for lipid_name in [row[0] for row in data]:
            for i in range(cycles):
                lipid = parser.parse(lipid_name)
                length += len(lipid_name)
                
                lipid_name2 = lipid.get_lipid_string(LipidLevel.STRUCTURE_DEFINED)
                lipid = parser.parse(lipid_name2)
                length += len(lipid_name2)
                
                lipid_name2 = lipid.get_lipid_string(LipidLevel.SN_POSITION)
                lipid = parser.parse(lipid_name2)
                length += len(lipid_name2)
                
                lipid_name2 = lipid.get_lipid_string(LipidLevel.SPECIES)
                lipid = parser.parse(lipid_name2)
                length += len(lipid_name2)
                
        elapsed = time.time() - start
        print("time elapsed: %f" % elapsed)
        print("lipid / second: %f" % (len(data) * cycles * 4 / elapsed))
        print("Avg. lipid name length: %f" % (length / (len(data) * cycles * 4)))
    """



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.Element import Element
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.LipidLevel import LipidLevel

class Cycle(FunctionalGroup):
    def __init__(self, cycle, start = None, end = None, double_bonds = None, functional_groups = None, bridge_chain = None):
        super().__init__("cy", functional_groups = functional_groups)
        self.count = 1
        self.cycle = cycle
        self.position = start if type(start) == int else -1
        self.start = start
        self.end = end
        self.double_bonds = double_bonds if double_bonds != None else 0
        self.elements[Element.H] = -2
        self.bridge_chain = bridge_chain if bridge_chain != None else []
        
        if type(start) != type(end):
            raise ConstraintViolationException("Cycle data start and end values not of same type!")
        
            
                
        
    def copy(self):
        cy = Cycle(cycle = self.cycle, start = self.start, end = self.end, bridge_chain = list(self.bridge_chain))
        cy.double_bonds = {k: v for k, v in self.double_bonds.items()} if type(self.double_bonds) == dict else self.double_bonds
        for fg, fg_list in self.functional_groups.items():
            cy.functional_groups[fg] = [func_group.copy() for func_group in fg_list]
        return cy
                
    
    
    def get_double_bonds(self):
        return super().get_double_bonds() + 1
    
    
    
    def rearrange_functional_groups(self, parent, shift):
        ## put everything back into parent
        if type(parent.double_bonds) != dict: parent.double_bonds = {}
        if type(self.double_bonds) == dict and len(self.double_bonds) > 0:
            for k, v in self.double_bonds.items(): parent.double_bonds[k] = v
            self.double_bonds = {}
        
        fgroup = parent.functional_groups
        for fg, fg_list in self.functional_groups.items():
            if fg not in fgroup: fgroup[fg] = []
            fgroup[fg] += fg_list
        self.functional_groups = {}
            
        # shift the cycle
        self.shift_positions(shift)
            
        ## take back what's mine
        # check double bonds
        if type(parent.double_bonds) == dict and len(parent.double_bonds) > 0:
            self.double_bonds = {db_pos: val for db_pos, val in parent.double_bonds.items() if type(self.start) == int and type(self.end) == int and self.start <= db_pos <= self.end}
            
            for pos in self.double_bonds:
                del parent.double_bonds[pos]
        
        # check functional groups
        remove_list = set()
        for fg, fg_list in fgroup.items():
            remove_item = []
            
            for i, func_group in enumerate(fg_list):
                if type(self.start) == int and type(self.end) == int and self.start <= func_group.position <= self.end and func_group != self:
                    if fg not in self.functional_groups: self.functional_groups[fg] = []
                    self.functional_groups[fg].append(func_group)
                    remove_item.append(i)
                    
            for i in remove_item[::-1]: del fgroup[fg][i]
            if len(fg_list) == 0: remove_list.add(fg)
            
        for fg in remove_list: del fgroup[fg]
        
        
    
    def shift_positions(self, shift):
        super().shift_positions(shift)
        if type(self.start) == int: self.start += shift
        if type(self.end) == int: self.end += shift
        if type(self.double_bonds) == dict:
            self.double_bonds = {k + shift: v for k, v in self.double_bonds.items()}
        
    
    
    def compute_elements(self):
        self.elements = {e: 0 for e in element_order}
        self.elements[Element.H] = -2
        if self.double_bonds != None:
            self.elements[Element.H] -= 2 * (self.double_bonds if type(self.double_bonds) == int else len(self.double_bonds))
            
        for chain_element in self.bridge_chain:
            if chain_element == Element.C:
                self.elements[Element.C] += 1
                self.elements[Element.H] += 2
                
            if chain_element == Element.N:
                self.elements[Element.N] += 1
                self.elements[Element.H] += 1
                
            if chain_element == Element.As:
                self.elements[Element.As] += 1
                self.elements[Element.H] += 1
                
            if chain_element == Element.P:
                self.elements[Element.P] += 1
                self.elements[Element.H] += 1
                
            if chain_element == Element.O:
                self.elements[Element.O] += 1
                
            if chain_element == Element.S:
                self.elements[Element.S] += 1
            
        # add all implicit carbon chain elements
        if self.start != None and self.end != None:
            n = max(0, self.cycle - (self.end - self.start + 1 + len(self.bridge_chain)))
            self.elements[Element.C] += n
            self.elements[Element.H] += 2 * n
            
        
        
    def to_string(self, level):
        cycle_string = ["["]
        if self.start != None and level == LipidLevel.FULL_STRUCTURE:
            cycle_string.append("%i-%i" % (self.start, self.end))
        
        if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE, LipidLevel.STRUCTURE_DEFINED} and len(self.bridge_chain) > 0:
            cycle_string.append("".join(element_shortcut[e] for e in self.bridge_chain))
        cycle_string.append("cy%i" % self.cycle)    
          
        if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE, LipidLevel.STRUCTURE_DEFINED}:
            if self.double_bonds != None:
                if type(self.double_bonds) != int:
                    cycle_string.append(":%i" % len(self.double_bonds))
                    db_positions = ["%i%s" % (k, self.double_bonds[k]) for k in sorted(self.double_bonds.keys())] if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE} else ["%i" % k for k in sorted(self.double_bonds.keys())]
                    db_pos = "(%s)" % ",".join(db_positions) if len (self.double_bonds) > 0 else ""
                    cycle_string.append(db_pos)
                else:
                    cycle_string.append(":%i" % self.double_bonds)
            
        
        if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE}:
            for fg in sorted(self.functional_groups.keys(), key = lambda x: x.lower()):
                fg_list = self.functional_groups[fg]
                fg_summary = ",".join([func_group.to_string(level) for func_group in sorted(fg_list, key = lambda x: x.position)])
                if len(fg_summary) > 0: cycle_string.append(";%s" % fg_summary)
        
        elif level == LipidLevel.STRUCTURE_DEFINED:
            for fg in sorted(self.functional_groups.keys()):
                fg_list = self.functional_groups[fg]
                if len(fg_list) > 0:
                    if len(fg_list) == 1:
                        fg_summary = ",".join([func_group.to_string(level) for func_group in fg_list])
                        if len(fg_summary) > 0: cycle_string.append(";%s" % fg_summary)
                
                    else:
                        fg_count = sum([func_group.count for func_group in fg_list])
                        if fg_count > 1: cycle_string.append(";(%s)%i" % (fg, fg_count))
                        else: cycle_string.append(";%s" % fg)
                    
        cycle_string.append("]")
        if level == LipidLevel.COMPLETE_STRUCTURE and self.stereochemistry != None: cycle_string.append("[%s]" % self.stereochemistry)
        
        return "".join(cycle_string)
        
        



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from pygoslin.domain.FattyAcid import FattyAcid
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.Element import *
from pygoslin.domain.LipidClass import all_lipids
from pygoslin.domain.FunctionalGroup import FunctionalGroup

ether_prefix = ["", "O-", "dO-", "tO-", "eO-"] 
        
class LipidSpeciesInfo(FattyAcid):
    
    def __init__(self, lipid_class):
        super().__init__("info")
        
        self.level = None
        self.num_ethers = 0
        self.num_specified_fa = 0
        self.poss_fa = all_lipids[lipid_class]["poss_fa"]
        self.total_fa = all_lipids[lipid_class]["max_fa"]
        self.extended_class = LipidFaBondType.ESTER
        self.lipid_class = lipid_class
        
        
        
    def copy(self):
        lsi = LipidSpeciesInfo(self.lipid_class)
        lsi.level = self.level
        lsi.num_ethers = self.num_ethers
        lsi.num_specified_fa = self.num_specified_fa
        lsi.position = self.position
        lsi.total_fa = self.total_fa
        lsi.extended_class = self.extended_class
        lsi.num_carbon = self.num_carbon
        lsi.double_bonds = {key: value for key, value in self.double_bonds.items()} if type(self.double_bonds) != int else self.double_bonds
        lsi.lipid_FA_bond_type = self.lipid_FA_bond_type
        
        for fg, fg_list in self.functional_groups.items():
            lsi.functional_groups[fg] = []
            for func_group in fg_list:
                lsi.functional_groups[fg].append(func_group.copy())
        
        return lsi
        
        
    def add(self, fa):
        
        if fa.lipid_FA_bond_type in {LipidFaBondType.ETHER_PLASMENYL, LipidFaBondType.ETHER_PLASMANYL} and fa.lipid_FA_bond_type not in {LipidFaBondType.LCB_REGULAR, LipidFaBondType.LCB_EXCEPTION}:
            self.num_ethers += 1
            self.lipid_FA_bond_type = LipidFaBondType.ETHER_PLASMANYL
            self.extended_class = fa.lipid_FA_bond_type
            
        elif fa.lipid_FA_bond_type in {LipidFaBondType.LCB_REGULAR, LipidFaBondType.LCB_EXCEPTION}:
            self.lipid_FA_bond_type = fa.lipid_FA_bond_type
        
                
        else:
            self.num_specified_fa += 1
        
        for fg, fg_list in fa.functional_groups.items():
            if fg not in self.functional_groups: self.functional_groups[fg] = []
            for func_group in fg_list:
                self.functional_groups[fg].append(func_group.copy())
         
        self.num_carbon += fa.get_elements()[Element.C]
        self.double_bonds += fa.get_double_bonds()
        
        
        
    def get_elements(self):
        elements = super().get_elements()
        if self.lipid_FA_bond_type != LipidFaBondType.LCB_EXCEPTION:
            elements[Element.O] -= (self.num_ethers == 0)
        elements[Element.H] += 1 if self.num_ethers == 0 else -1
        
        return elements
        
    
    def to_string(self):
        global ether_prefix
        
        info_string = [ether_prefix[self.num_ethers]]
        info_string.append("%i:%i" % (self.num_carbon, self.double_bonds))
        

        fg_elements = self.get_functional_group_elements()
        additional_elements = ";".join("%s%s" % (element_shortcut[e], str(fg_elements[e]) if fg_elements[e] > 1 else "") for e in element_order[2:] if e in fg_elements and fg_elements[e] > 0)
        
        if len(additional_elements) > 0:
            info_string.append(";")
            info_string.append(additional_elements)
            
        return "".join(info_string)



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pygoslin
from pygoslin.domain.Element import *
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.parser.ParserCommon import SumFormulaParser, Parser
from os import path

class FunctionalGroup:
    def __init__(self, name, position = -1, count = 1, double_bonds = 0, stereochemistry = None, elements = None, functional_groups = None, is_atomic = False, stereo_bound = False):
        self.name = name
        self.position = position
        self.count = count
        self.stereochemistry = stereochemistry
        self.stereo_bound = stereo_bound
        self.ring_stereo = ""
        self.double_bonds = double_bonds
        self.is_atomic = is_atomic
        self.elements = {e: 0 for e in Element} if elements == None else {k: v for k, v in elements.items()}
        self.functional_groups = functional_groups if functional_groups != None else {}
        
        
    def copy(self):
        functional_group = FunctionalGroup(self.name, position = self.position, count = self.count, double_bonds = self.double_bonds, stereochemistry = self.stereochemistry, elements = {k: v for k, v in self.elements.items()}, is_atomic = self.is_atomic, stereo_bound = self.stereo_bound)
        for fg, fg_list in self.functional_groups.items():
            if fg not in functional_group.functional_groups: functional_group.functional_groups[fg] = []
            for func_group in fg_list:
                functional_group.functional_groups[fg].append(func_group.copy())
        functional_group.ring_stereo = self.ring_stereo
        return functional_group
        
        
        
    def get_elements(self):
        self.compute_elements()
        elements = {e: 0 for e in Element}
        for k, v in self.elements.items(): elements[k] = v
        for k, v in self.get_functional_group_elements().items(): elements[k] += v
        return elements
    
    
    
    def shift_positions(self, shift):
        self.position += shift
        for fg_name, fg_list in self.functional_groups.items():
            for fg in fg_list:
                fg.shift_positions(shift)
                
                
                
    def stereo_information_missing(self):
        missing = self.stereo_bound and self.stereochemistry == ""
        for fg_name, fg_list in self.functional_groups.items():
            for fg in fg_list:
                missing |= fg.stereo_information_missing()
        return missing
    
    
    def get_functional_group_elements(self):
        elements = {e: 0 for e in Element}
        
        for fg, fg_list in self.functional_groups.items():
            for func_group in fg_list:
                for k, v in func_group.get_elements().items():
                    if k not in elements: elements[k] = 0
                    elements[k] += v * func_group.count
                    
        return elements
    
    
    
    def compute_elements(self):
        for fg, fg_list in self.functional_groups.items():
            for func_group in fg_list:
                func_group.compute_elements()
        
        
        
    def add_functional_group(self, func_group):
        if func_group.name not in self.functional_groups: self.functional_groups[func_group.name] = []
        self.functional_groups[func_group.name].append(func_group)
        
        
    def to_string(self, level):
        fg_string = ""
        if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE}:
            if str.isnumeric(self.name[0]): fg_string = "%i%s(%s)" % (self.position, self.ring_stereo, self.name) if self.position > -1 else self.name
            else: fg_string = "%i%s%s" % (self.position, self.ring_stereo, self.name) if self.position > -1 else self.name
            
        else:
            fg_string = "(%s)%i" % (self.name, self.count) if self.count > 1 else self.name
        if self.stereochemistry != None and self.stereochemistry != "" and level == LipidLevel.COMPLETE_STRUCTURE: fg_string += "[%s]" % self.stereochemistry
                
        return fg_string
    
    
    
    def get_total_functional_group_count(self, fg_name):
        return sum([fg_item.count for fg_item in self.functional_groups[fg_name]]) if fg_name in self.functional_groups else 0
    
    
    
    def get_double_bonds(self):
        db = self.count * (self.double_bonds if type(self.double_bonds) == int else len(self.double_bonds))
        for fg, fg_list in self.functional_groups.items():
            for func_group in fg_list:
                db += func_group.get_double_bonds()
                
        return db
    
    
    def __iadd__(self, fgroup):
        for e in Element:
            if e not in self.elements: self.elements[e] = 0
        
        for k, v in fgroup.get_elements().items():
            self.elements[k] += v * fgroup.count
        
        return self




    
class HeadgroupDecorator(FunctionalGroup):
    def __init__(self, name, position = -1, count = 1, elements = None, suffix = False, level = None):
        super().__init__(name, position = position, count = count, elements = elements)
        self.suffix = suffix
        self.lowest_visible_level = level
        
    def copy(self):
        return HeadgroupDecorator(self.name, position = self.position, count = self.count, elements = self.elements, suffix = self.suffix, level = self.lowest_visible_level)
        
        
        
        
    def to_string(self, level):
        if not self.suffix: return self.name + ("%i" % self.count if self.count > 1 else "")
    
        decorator_string = ""
        if self.lowest_visible_level == None or self.lowest_visible_level.value <= level.value:
            if "decorator_alkyl" in self.functional_groups:
                if len(self.functional_groups["decorator_alkyl"]) > 0:
                    decorator_string = self.functional_groups["decorator_alkyl"][0].to_string(level) if level.value > LipidLevel.SPECIES.value else "Alk"
                else:
                    decorator_string = "Alk"
                
            elif "decorator_acyl" in self.functional_groups:
                if len(self.functional_groups["decorator_acyl"]) > 0:
                    decorator_string = "FA %s" % self.functional_groups["decorator_acyl"][0].to_string(level) if level.value > LipidLevel.SPECIES.value else "FA"
                else:
                    decorator_string = "FA"
                
            else:
                decorator_string = self.name
                
            decorator_string = "(%s)" % decorator_string
            
        return decorator_string
    
    



    
class AcylAlkylGroup(FunctionalGroup):
    def __init__(self, fa, position = -1, count = 1, alkyl = False, N_bond = False):
        
        super().__init__("O", position = position, count = count)
        self.alkyl = alkyl
        if fa != None: self.functional_groups["alkyl" if self.alkyl else "acyl"] = [fa]
        self.double_bonds = int(not self.alkyl)
        self.set_N_bond_type(N_bond)
        
        if self.N_bond:
            self.elements[Element.H] = 2 if self.alkyl else 0
            self.elements[Element.O] = -1 if self.alkyl else 0
            self.elements[Element.N] = 1
            
        else:
            self.elements[Element.H] = 1 if self.alkyl else -1
            self.elements[Element.O] = 0 if self.alkyl else 1
        
        
    def copy(self):
        return AcylAlkylGroup(self.functional_groups["alkyl" if self.alkyl else "acyl"][0].copy(), alkyl = self.alkyl, position = self.position, count = self.count, N_bond = self.N_bond)
        
        
        
    def set_N_bond_type(self, N_bond):
        self.N_bond = N_bond
        
        if self.N_bond:
            self.elements[Element.H] = 2 if self.alkyl else 0
            self.elements[Element.O] = -1 if self.alkyl else 0
            self.elements[Element.N] = 1
            
        else:
            self.elements[Element.H] = 1 if self.alkyl else -1
            self.elements[Element.O] = 0 if self.alkyl else 1
        
        

    def to_string(self, level):
        acyl_alkyl_string = []
        if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE}: acyl_alkyl_string.append("%i" % self.position)
        acyl_alkyl_string.append("%s(" % ("N" if self.N_bond else "O"))
        if not self.alkyl: acyl_alkyl_string.append("FA ")
        fa = self.functional_groups["alkyl" if self.alkyl else "acyl"][0]
        acyl_alkyl_string.append(fa.to_string(level))
        acyl_alkyl_string.append(")")
        
        return "".join(acyl_alkyl_string)
    
    



    
class CarbonChain(FunctionalGroup):
    def __init__(self, fa, position = -1, count = 1):
        super().__init__("cc", position = position, count = count)
        if fa != None: self.functional_groups["cc"] = [fa]
        
        self.elements[Element.H] = 1
        self.elements[Element.O] = -1
        
        
    def copy(self):
        return CarbonChain(self.functional_groups["cc"][0].copy(), position = self.position, count = self.count)
        
        

    def to_string(self, level):
        return "%s(%s)" % ((str(self.position) if level in {LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE} else ""), self.functional_groups["cc"][0].to_string(level))
    
    
    
_known_functional_groups, s = {}, SumFormulaParser()
fg_dir_name = path.dirname(pygoslin.__file__)
fg_file_name = path.join(fg_dir_name, "data", "goslin", "functional-groups.csv")
with open(fg_file_name, mode = "rt", encoding= "utf-8") as fg_infile:
    fg_infile.readline()
    for line in fg_infile:
        line = line.strip()
        row = Parser.split_string(line.strip(), ",", '"', True)
        e = s.parse(row[2]) if len(row[2]) > 0 else {e: 0 for e in Element}
        row.append(row[1])
        for i in range(7, len(row)):
            key = row[i]
            if len(key) == 0: continue
        
            if row[0] == "FG":
                _known_functional_groups[key] = FunctionalGroup(row[1], elements = e, double_bonds = int(row[3]), is_atomic = row[4] == "1", stereo_bound = row[5] == "1")
                
            elif row[0] == "HGD":
                _known_functional_groups[key] = HeadgroupDecorator(row[1], elements = e)
        
    

def get_functional_group(name):
    if name in _known_functional_groups:
        return _known_functional_groups[name].copy()
    raise Exception("Name '%s' not registered in functional group list" % name)



        
        
        



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.domain.LipidExceptions import RuntimeException
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.LipidCategory import LipidCategory
from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.LipidSpeciesInfo import LipidSpeciesInfo
from pygoslin.domain.LipidClass import *
from pygoslin.domain.LipidFaBondType import *
from pygoslin.domain.HeadGroup import HeadGroup
from pygoslin.domain.Element import Element
from pygoslin.domain.FattyAcid import FattyAcid

class LipidSpecies:
    def __init__(self, headgroup, fa = []):
        self.headgroup = headgroup
        
        self.info = LipidSpeciesInfo(self.headgroup.lipid_class)
        self.info.level = LipidLevel.SPECIES
        self.fa_list = []
        
        # add fatty acids and count numbers for fatty acyl chains
        fa_it = len(fa) > 0 and fa[0].lipid_FA_bond_type in {LipidFaBondType.LCB_EXCEPTION, LipidFaBondType.LCB_REGULAR}
        for i, fatty_acid in enumerate(fa):
            if i > 0 or not fa_it: fatty_acid.name += "%i" % (i + 1 - fa_it)
            fatty_acid.position = -1
            self.info.add(fatty_acid)
            
        
    def get_extended_class(self):
        special_case = self.headgroup.lipid_category == LipidCategory.GP if self.info != None and self.info.num_carbon > 0 else False
        if special_case and self.info.extended_class in [LipidFaBondType.ETHER_PLASMANYL, LipidFaBondType.ETHER_UNSPECIFIED]:
            return all_lipids[self.headgroup.lipid_class]["name"] + "-O"
        
        if special_case and self.info.extended_class == LipidFaBondType.ETHER_PLASMENYL:
            return all_lipids[self.headgroup.lipid_class]["name"] + "-P"
        
        return all_lipids[self.headgroup.lipid_class]["name"]
            


    def get_lipid_string(self, level = None):        
        
        if level in {LipidLevel.CATEGORY, LipidLevel.CLASS}:
            return self.headgroup.get_lipid_string(level)
        
        elif level == None or level == LipidLevel.SPECIES:
            lipid_string = [self.headgroup.get_lipid_string(level)]
            
            if self.info != None and (self.info.elements[Element.C] > 0 or self.info.num_carbon > 0):
                lipid_string += " " if all_lipids[self.headgroup.lipid_class]["category"] != LipidCategory.ST else "/"
                a = self.info.copy()
                for decorator in self.headgroup.decorators:
                    if decorator.name in {"decorator_alkyl", "decorator_acyl"}:
                        a.num_carbon += decorator.get_elements()[Element.C]
                        a.double_bonds += decorator.get_double_bonds()
                
                lipid_string += a.to_string()
                
            return "".join(lipid_string)
        
        else:
            raise RuntimeException("LipidSpecies does not know how to create a lipid string for level %s" % level)
        
        
        
        
    def get_elements(self):
        if self.info.level not in {LipidLevel.STRUCTURE_DEFINED, LipidLevel.COMPLETE_STRUCTURE, LipidLevel.FULL_STRUCTURE, LipidLevel.SPECIES, LipidLevel.MOLECULAR_SPECIES, LipidLevel.SN_POSITION} or self.headgroup.use_headgroup:
            raise LipidException("Element table cannot be computed for lipid level '%s'" % self.info.level)
        
        dummy = FunctionalGroup("dummy", elements = self.headgroup.get_elements())
        dummy += self.info
        
        # since only one FA info is provided, we have to treat this single information as
        # if we would have the complete information about all possible FAs in that lipid
        additional_fa = all_lipids[self.headgroup.lipid_class]["poss_fa"]
        remaining_H = all_lipids[self.headgroup.lipid_class]["max_fa"] - additional_fa
        hydrochain = int("HC" in all_lipids[self.headgroup.lipid_class]["specials"])
        dummy.elements[Element.O] -= -additional_fa + self.info.num_ethers + self.headgroup.sp_exception + hydrochain
        dummy.elements[Element.H] += -additional_fa + remaining_H + 2 * self.info.num_ethers + 2 * hydrochain
        
        if "Amide" in all_lipids[self.headgroup.lipid_class]["specials"]:
            dummy.elements[Element.O] -= all_lipids[self.headgroup.lipid_class]["max_fa"]
            dummy.elements[Element.H] += all_lipids[self.headgroup.lipid_class]["max_fa"]
        
        return dummy.elements
        
        
        
    def sort_fatty_acyl_chains(self):
        pass



"""
MIT License

Copyright (c) 2020 Dominik Kopczynski   -   dominik.kopczynski {at} univie.ac.at
                   Nils Hoffmann  -  nils.hoffmann {at} cebitec.uni-bielefeld.de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from pygoslin.domain.FunctionalGroup import *
from pygoslin.domain.LipidExceptions import *
from pygoslin.domain.Element import Element
from pygoslin.domain.LipidFaBondType import LipidFaBondType
from pygoslin.domain.LipidLevel import LipidLevel
from pygoslin.domain.FunctionalGroup import *

class FattyAcid(FunctionalGroup):
    LCB_STATES = set([LipidFaBondType.LCB_REGULAR, LipidFaBondType.LCB_EXCEPTION])

    def __init__(self, name, num_carbon = 0, double_bonds = 0, functional_groups = None, lipid_FA_bond_type = LipidFaBondType.ESTER, position = 0):
        super().__init__(name, double_bonds = double_bonds, position = position, functional_groups = functional_groups)
        self.num_carbon = num_carbon
        self.lipid_FA_bond_type = lipid_FA_bond_type
        self.unresolved_hidden_fa = False
        
        if lipid_FA_bond_type == LipidFaBondType.LCB_REGULAR: self.functional_groups["[X]"] = [get_functional_group("X")]
        
        num_double_bonds = len(self.double_bonds) if type(self.double_bonds) != int else self.double_bonds
        if num_carbon < 0 or num_carbon == 1:
            raise ConstraintViolationException("FattyAcid must have at least 2 carbons!")
        
        if num_double_bonds < 0:
            raise ConstraintViolationException("FattyAcid must have at least 0 double bonds!")
        
        
        
    def copy(self):
        fa = FattyAcid(self.name)
        fa.position = self.position
        fa.num_carbon = self.num_carbon
        fa.lipid_FA_bond_type = self.lipid_FA_bond_type
        fa.double_bonds = {key: value for key, value in self.double_bonds.items()} if type(self.double_bonds) != int else self.double_bonds
        for fg, fg_list in self.functional_groups.items():
            fa.functional_groups[fg] = [func_group.copy() for func_group in fg_list]
        return fa

        
        
        
    def clone(self, fa):
        self.name = fa.name
        self.position = fa.position
        self.num_carbon = fa.num_carbon
        self.lipid_FA_bond_type = fa.lipid_FA_bond_type
        self.double_bonds = {key: value for key, value in fa.double_bonds.items()} if type(fa.double_bonds) != int else fa.double_bonds
        self.functional_groups = {}
        for fg, fg_list in fa.functional_groups.items():
            self.functional_groups[fg] = []
            for fg_item in fg_list:
                func_group = FunctionalGroup("")
                func_group.clone(fg_item)
                self.function_groups[fg].append(func_group.copy())
                
    
    
    def set_type(self, lipid_FA_bond_type):
        self.lipid_FA_bond_type = lipid_FA_bond_type
        if self.lipid_FA_bond_type == LipidFaBondType.LCB_REGULAR and "[X]" not in self.functional_groups:
            self.functional_groups["[X]"] = [get_functional_group("X")]
            
        elif "[X]" in self.functional_groups:
            del self.functional_groups["[X]"]
            
        self.name = "FA" if self.lipid_FA_bond_type not in {LipidFaBondType.LCB_REGULAR, LipidFaBondType.LCB_EXCEPTION} else "LCB"
                
                
    def get_double_bonds(self):
        return super().get_double_bonds() + (self.lipid_FA_bond_type == LipidFaBondType.ETHER_PLASMENYL)
        
                
    def db_num(self):
        return self.double_bonds if type(self.double_bonds) == int else len(self.double_bonds)
                
        
    def to_string(self, level):
        fa_string = [self.lipid_FA_bond_type.prefix()]
        num_carbon = self.num_carbon
        double_bonds = self.double_bonds
        
        if num_carbon == 0 and double_bonds == 0 and level not in {LipidLevel.STRUCTURE_DEFINED, LipidLevel.FULL_STRUCTURE, LipidLevel.COMPLETE_STRUCTURE, LipidLevel.SN_POSITION}:
            return ""
        
        if level in {LipidLevel.MOLECULAR_SPECIES, LipidLevel.SN_POSITION}:
            num_carbon = self.get_elements()[Element.C]
            double_bonds = self.get_double_bonds() - (self.lipid_FA_bond_type == LipidFaBondType.ETHER_PLASMENYL)
            

        fa_string.append("%i" % num_carbon)
        
        if type(double_bonds) != int:
            fa_string.append(":%i" % len(double_bonds))
            if level in {LipidLevel.FULL_STRUCTURE, LipidLevel.COMPLETE_STRUCTURE}:
                db_positions = ["%i%s" % (k, double_bonds[k]) for k in sorted(double_bonds.keys())]
                db_pos = "(%s)" % ",".join(db_positions) if len (double_bonds) > 0 else ""
                fa_string.append(db_pos)
            elif level == LipidLevel.STRUCTURE_DEFINED:
                db_positions = ["%i" % k for k in sorted(double_bonds.keys())]
                db_pos = "(%s)" % ",".join(db_positions) if len (double_bonds) > 0 else ""
                fa_string.append(db_pos)
            
        else:
            fa_string.append(":%i" % double_bonds)
            
            
        if level == LipidLevel.COMPLETE_STRUCTURE and self.stereochemistry not in {None, ""}:
            fa_string.append("[%s]" % self.stereochemistry)
            
        
        if level in {LipidLevel.FULL_STRUCTURE, LipidLevel.COMPLETE_STRUCTURE}:
            for fg in sorted(self.functional_groups.keys(), key = lambda x: x.lower()):
                if fg == "[X]": continue
                fg_list = self.functional_groups[fg]
                fg_summary = ",".join([func_group.to_string(level) for func_group in sorted(fg_list, key = lambda x: x.position)])
                if len(fg_summary) > 0: fa_string.append(";%s" % fg_summary)
        
        elif level == LipidLevel.STRUCTURE_DEFINED:
            for fg in sorted(self.functional_groups.keys(), key = lambda x: x.lower()):
                if fg == "[X]": continue
                fg_list = self.functional_groups[fg]
                if len(fg_list) > 0:
                    
                    if fg in {"acyl", "alkyl", "cy", "cc", "acetoxy"}:
                        fg_summary = ",".join([func_group.to_string(level) for func_group in fg_list])
                        if len(fg_summary) > 0: fa_string.append(";%s" % fg_summary)
                    
                    else:
                        fg_count = sum([func_group.count for func_group in fg_list])
                        if fg_count > 1: fa_string.append(";(%s)%i" % (fg, fg_count) if not fg_list[0].is_atomic else ";%s%i" % (fg, fg_count))
                        else: fa_string.append(";%s" % fg)
        
        
        else:
            elements = self.get_functional_group_elements()
            additional_elements = ";".join(("%s%i" % (element_shortcut[e], elements[e]) if elements[e] > 1 else "%s" % element_shortcut[e]) for e in element_order[2:] if e in elements and elements[e] > 0)
            
            if len(additional_elements) > 0:
                fa_string.append(";")
                fa_string.append(additional_elements)
        
        return "".join(fa_string)
    
    

        
    
    def get_functional_group_elements(self):
        elements = super().get_functional_group_elements()
        # subtract the invisible [X] functional group for regular LCBs
        if self.lipid_FA_bond_type == LipidFaBondType.LCB_REGULAR and "O" in self.functional_groups:
            elements[Element.O] -= 1 
        return elements
    
    
    def num_oxygens(self):
        return self.get_functional_group_elements()[Element.O]
    
    

    def compute_elements(self):
        self.elements = {e: 0 for e in Element}
        if self.unresolved_hidden_fa:
            self.elements[Element.O] = 1
            self.elements[Element.H] = -1
            return self.elements
        
        num_double_bonds = len(self.double_bonds) if type(self.double_bonds) != int else self.double_bonds
        if self.lipid_FA_bond_type == LipidFaBondType.ETHER_PLASMENYL: num_double_bonds += 1
        
        if self.num_carbon == 0 and num_double_bonds == 0:
            self.elements[Element.H] = 1
            return
        
        if self.lipid_FA_bond_type not in {LipidFaBondType.LCB_REGULAR, LipidFaBondType.LCB_EXCEPTION}:
            
            self.elements[Element.C] = self.num_carbon # carbon
            if self.lipid_FA_bond_type == LipidFaBondType.ESTER:
                self.elements[Element.H] = (2 * self.num_carbon - 1 - 2 * num_double_bonds) # hydrogen
                self.elements[Element.O] = 1 # oxygen
            
            elif self.lipid_FA_bond_type == LipidFaBondType.ETHER_PLASMENYL:
                self.elements[Element.H] = (2 * self.num_carbon - 1 - 2 * num_double_bonds + 2) # hydrogen
            
            elif self.lipid_FA_bond_type in {LipidFaBondType.ETHER_PLASMANYL, LipidFaBondType.ETHER}:
                self.elements[Element.H] = ((self.num_carbon + 1) * 2 - 1 - 2 * num_double_bonds) # hydrogen
            
            elif self.lipid_FA_bond_type == LipidFaBondType.AMIDE:
                self.elements[Element.H] = (2 * self.num_carbon + 1 - 2 * num_double_bonds) - 1 # hydrogen
                
            else:
                raise LipidException("Mass cannot be computed for fatty acyl chain with bond type: %s" % self.lipid_FA_bond_type)
                
        else:
            # long chain base
            self.elements[Element.C] = self.num_carbon # carbon
            self.elements[Element.H] = (2 * (self.num_carbon - num_double_bonds) + 1) # hydrogen
            self.elements[Element.N] = 1 # nitrogen
            #self.elements[Element.O] = 1 # oxygen
            
            
